//------------------------------------------------------------------------------
//
//  scene.cpp
//
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//  include
//------------------------------------------------------------------------------
#include "common.h"

#include <vector>

//------------------------------------------------------------------------------
//  Random(ĂƂ)
//------------------------------------------------------------------------------
int random() {
  static int a = timeGetTime();
  //static int a = 1;
  static int b = 8345678;
  static int c = 8012345;
  a += b; b += c; c += a;
  return (int)(a >> 16);
}

float frandom() {
  return (float)random() / (FLOAT)0x7FFF;
}


//------------------------------------------------------------------------------
//  Line Font
//------------------------------------------------------------------------------
class D3DLineFont {
  int  Height;
  float Dist;
  float shadediv;
  char FaceName[LF_FACESIZE];

  struct D3D_LINE {
    D3DXVECTOR4 pos;
    D3DCOLOR    color;
    enum { FVF = D3DFVF_XYZRHW | D3DFVF_DIFFUSE };
  };

  struct FontVertexData {
    UINT  code;
    FLOAT fRight;
    std::vector<D3D_LINE> vertex;
  };

  std::vector<FontVertexData> vData;
  int is2byte(unsigned char *s) {
    unsigned char c = s[0];
    if ( (c >= 0x81 && c <= 0x9f) || (c >= 0xe0 && c <= 0xfc)) {
      c = s[1];
      if (c >= 0x40 && c <= 0xfc) return 1;
    }
    return 0;
  }

  FLOAT fix2float(const FIXED& fixed) {
    return FLOAT(fixed.value) + FLOAT(fixed.fract) / 65536.0;
  }

  void PushVertexFVD(FontVertexData *fvd, TEXTMETRIC *tm, GLYPHMETRICS *gm, float a, float b, int is_chain) {
    //b += tm->tmAscent - gm->gmptGlyphOrigin.y;
    D3D_LINE temp = {
      D3DXVECTOR4(a, b, 0, 1), 0xFFFFFFFF
    };
    fvd->vertex.push_back(temp);
    if(!(fvd->vertex.size() & 1) && is_chain) {
      D3D_LINE temp = {
        D3DXVECTOR4(a, b, 0, 1), 0xFFFFFFFF
      };
      fvd->vertex.push_back(temp);
    }
  }

  HRESULT MakeFontVertex(FontVertexData *fvd, UINT Height, UINT code) {
    UINT targetChar = code;
    HDC   hdc  = NULL;
    HFONT font = NULL;
    MAT2  mat2 = {{0, 1}, {0, 0}, {0, 0}, {0, -1}};
    GLYPHMETRICS gm;
    TEXTMETRIC   tm;
    LOGFONT logfont = {
      Height, 0, 0, logfont.lfEscapement, FW_NORMAL, FALSE, FALSE, FALSE,
      SHIFTJIS_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
      DEFAULT_QUALITY, FF_DONTCARE|DEFAULT_PITCH
    };
    wsprintf(logfont.lfFaceName, "%s", FaceName);

    hdc  = CreateCompatibleDC(NULL);
    font = CreateFontIndirect(&logfont);
    SelectObject(hdc, font);
    GetTextMetrics(hdc, &tm);
    
    DWORD rSize = GetGlyphOutline(hdc, targetChar, GGO_BEZIER | GGO_NATIVE | GGO_UNHINTED, &gm, 0, NULL, &mat2 );
    if(rSize != GDI_ERROR) {
      char *buff = new char[rSize * 2];
      GetGlyphOutline(hdc, targetChar, GGO_BEZIER|GGO_NATIVE|GGO_UNHINTED, &gm, rSize, buff, &mat2 );
      const char* p = buff;
      int bailout = 0;
      while (p < buff + rSize && !bailout) {
        const TTPOLYGONHEADER* pHeader = (const TTPOLYGONHEADER *)p;
        p += sizeof(TTPOLYGONHEADER);
        PushVertexFVD(fvd,  &tm, &gm, fix2float(pHeader->pfxStart.x), fix2float(pHeader->pfxStart.y), 0);
        while (p < (char*)pHeader + (pHeader->cb) &&  !bailout) {
          const TTPOLYCURVE* pCurve = (const TTPOLYCURVE *)p;
          switch(pCurve->wType) {
          case TT_PRIM_LINE:
            for (WORD i = 0; i < pCurve->cpfx; i++) PushVertexFVD(fvd,  &tm, &gm, fix2float(pCurve->apfx[i].x), fix2float(pCurve->apfx[i].y), 1);
            break;
          case TT_PRIM_QSPLINE:
            for (WORD i = 0; i < pCurve->cpfx; i++) { }
            break;
          case TT_PRIM_CSPLINE:
            for (WORD i = 0; i < pCurve->cpfx; i++) PushVertexFVD(fvd, &tm, &gm, fix2float(pCurve->apfx[i].x), fix2float(pCurve->apfx[i].y), 1);
            break;
          default:
            printf("%s : Error", __FUNCTION__);
            bailout = 1;
            break;
          }
          p += (sizeof(WORD)*2) + (sizeof(POINTFX) * pCurve->cpfx);
        }
        PushVertexFVD(fvd,  &tm, &gm, fix2float(pHeader->pfxStart.x), fix2float(pHeader->pfxStart.y), 0);
      }
      delete [] buff;
    } else {
      printf("GDI Error\n");
    }
    DeleteObject(font);
    DeleteDC(hdc);
    fvd->fRight = gm.gmCellIncX;
    return S_OK;
  }
public:
  D3DLineFont() : Height(32) {
    wsprintf(FaceName, "");
    Dist = 0;
    shadediv = 32;
  }

  ~D3DLineFont() {}

  HRESULT Setup(LPCSTR name, UINT Height) {
    HRESULT hRet = E_FAIL;
    if(name) {
      wsprintf(FaceName, name);
      this->Height = Height;  
      hRet = S_OK;
    }
    return hRet;
  }

  HRESULT SetDistortion(FLOAT f) {
    Dist = f;
    return 0;
  }

  HRESULT Draw(char *textdata, FLOAT x, FLOAT y, D3DCOLOR color) {
    unsigned char *data = (unsigned char *)textdata;
    D3DXVECTOR4 pos     = D3DXVECTOR4(x, y, 0, 0);
    while(*data) {
      FontVertexData fvd;
      fvd.code = 0;
      UINT code = 0;
      
      //is 2byte code?
      if(is2byte(data)) {
        code = (data[0] << 8) | (data[1]);
        data += 2;
      } else {
        code = *data;
        data++;
        if(!isgraph(code)) {
          if(isspace(code)) {
            pos.x += Height/2;
          }
          continue;
        }
      }
      
      //scan code pool.
      {
        std::vector<FontVertexData>::iterator ite = vData.begin();
        while(ite != vData.end()) {
          if(ite->code == code) {
            //Copy
            fvd = *ite;
            break;
          }
          ite++;
        }
      }
      
      //if not found code then make font vertex.
      if(fvd.code == 0) {
        fvd.code = code;
        if(MakeFontVertex(&fvd, Height, fvd.code) == S_OK) {
          vData.push_back(fvd);
        } else {
          return E_FAIL;
        }
      }
      
      //draw
      {
        //Translate position.
        FontVertexData temp = fvd;
        std::vector<D3D_LINE>::iterator ite;
        ite = temp.vertex.begin();
        while(ite != temp.vertex.end()) {
          ite->pos += pos;
          ite->color = color;
          float a = sin( (ite->pos.x - Height / 2) * D3DX_PI );
          float b = cos( (ite->pos.x - Height / 2) * D3DX_PI );
          ite->pos.x += (a) * Dist + frandom();
          ite->pos.y += (b) * Dist + frandom();
          ite++;
        }
        
        
        d3ddevice->SetFVF(D3D_LINE::FVF);
        //Shade
        if(shadediv) {
          FontVertexData temp2 = temp;
          ite = temp2.vertex.begin();
          while(ite != temp2.vertex.end()) {
            ite->color = color & 0xFF7F7F7F;
            ite->pos.x += Height / shadediv;
            ite->pos.y += Height / shadediv;
            ite++;
          }
          d3ddevice->DrawPrimitiveUP(D3DPT_LINELIST, temp2.vertex.size() / 2, &temp2.vertex[0], sizeof(D3D_LINE));
        }
        d3ddevice->DrawPrimitiveUP(D3DPT_LINELIST, temp.vertex.size() / 2, &temp.vertex[0], sizeof(D3D_LINE));
      }
      pos.x += fvd.fRight;
    }
    return S_OK;
  }
};


D3DLineFont lf;


//------------------------------------------------------------------------------
//
//  function definition
//
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//  InitScene
//------------------------------------------------------------------------------
void InitScene() {

  TermScene();

  //d3ddevice->SetRenderState(D3DRS_CULLMODE,  D3DCULL_NONE);
  //d3ddevice->SetRenderState(D3DRS_ZENABLE, FALSE);
  d3ddevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  d3ddevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  d3ddevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);
  d3ddevice->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
  d3ddevice->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
}

//------------------------------------------------------------------------------
//  TermScene
//------------------------------------------------------------------------------
void TermScene() {
}

//------------------------------------------------------------------------------
//  DoScene
//------------------------------------------------------------------------------
void DoScene() {
  //d3ddevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
  //d3ddevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
  static float dist = 0;
  static int count = 0;
  count++;
  if(GetAsyncKeyState(VK_F5) || !(count % 500)) {
    dist = 200;
  }
  if(GetAsyncKeyState('Z')) {
    dist = sin(count * 3.141592 / 180) * 3;
  }
  if(GetAsyncKeyState('A')) {
    dist += 0.1;
  }
  dist *= 0.9;
  lf.SetDistortion(dist);
  
  d3ddevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
  d3ddevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
  d3ddevice->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL, 0x000000, 1.0f, 0);
  if(d3ddevice->BeginScene() == D3D_OK) {
    int H = 70;
    lf.Setup("Courier", H);
    int Line = H;
    lf.Draw("AEgCœ{", 0, Line, 0xFF00FF00);
    Line += H;
    lf.Draw("\I", 0, Line, 0xFF00FF00);
    Line += H;
    lf.Draw("ABCDEFGHIJKLMNOPQRSTUVWXYZ", 0, Line, 0xFFFF0000);
    Line += H;
    lf.Draw("abcdefghijklmnopqrstuvwxyz", 0, Line, 0xFF00FF00);
    Line += H;
    lf.Draw("!\"#$%&'()", 0, Line, 0xFF00FF00);
    Line += H;
    Line += H;
    Line += H;
    Line += H;
    lf.Draw("F5L[łɂ܂", 0, Line, 0xFFFFFFFF);
    
    char temp[123];
    sprintf_s(temp, "%d %lf\n", count, dist);
    Line += H;
    lf.Draw(temp, 0, Line, 0xFF77FFee);
    lf.Setup("Courier", 256);
    lf.Draw("鳖", 0, 512, 0xFF00FFFF);
    d3ddevice->EndScene();
  }
  d3ddevice->Present( NULL, NULL, NULL, NULL );
}

